# Copyright (c) 2024, SMLIGHT <smartlight.email@gmail.com>
# SPDX-License-Identifier: Apache-2.0

import binascii
from dataclasses import dataclass, field
from io import StringIO
import logging
import struct
from typing import Any, List

from intelhex import IntelHex
import magic

from .exceptions import FwException

_LOGGER = logging.getLogger(__name__)


@dataclass
class Segment:
    start: int
    end: int
    size: int = field(init=False)
    bytes: bytearray

    def __post_init__(self):
        self.size = self.end - self.start

        # remove the last byte, which is the empty
        pop = self.bytes.pop()
        assert pop == 255, "Segment data does not end with 0xFF"

        assert len(self.bytes) == self.size, "Segment size does not match data size"
        self.crc32 = binascii.crc32(bytearray(self.bytes)) & 0xFFFFFFFF
        _LOGGER.debug("Segment: 0x%08x, %d bytes", self.start, self.size)


class FirmwareFile:
    def __init__(self, *, path: str | None = None, buffer: bytearray | None = None):
        """
        Read a firmware file and store its data ready for device programming.

        This class will try to guess the file type using python-magic.

        If python-magic indicates a plain text file, then the file will be
        treated as one of Intel HEX format. In all other cases, the file will
        be treated as a raw binary file.

        In both cases, the file's contents are stored in bytes for subsequent
        usage to program a device or to perform a crc check.

        Parameters:
            path -- A str with the path to the firmware file.

        Attributes:
            bytes: A bytearray with firmware contents ready to send to the
            device
        """
        self._crc32: int | None = None
        self.segments: List[Any] = []
        self.ih: IntelHex
        self.size: int = 0
        file_type = None

        if path:
            file_type = magic.from_file(path, mime=True)
        elif buffer:
            file_type = magic.from_buffer(bytes(buffer), mime=True)
        else:
            return

        if file_type == "text/plain":
            _LOGGER.info("Firmware file: Intel Hex")

            fobj = StringIO(bytearray(buffer).decode()) if buffer else path
            self.read_hex(fobj)
        elif (
            file_type == "application/octet-stream"
            or file_type == "application/x-dosexec"
        ):
            _LOGGER.info("Firmware file: Raw Binary")
            self._read_bin(path)
        else:
            error_str = f"Could not determine firmware type, {file_type}"
            raise FwException(error_str)

    def read_hex(self, path: str | StringIO):
        self.ih = IntelHex(path)
        self._process_hex()

    async def from_buffer(self, buffer):
        blob = await buffer.text()
        fobj = StringIO(blob)
        self.ih = IntelHex(fobj)
        self._process_hex()

    def _process_hex(self):
        for segment in self.ih.segments(min_gap=2048):
            data = bytearray(self.ih.tobinarray(*segment))
            self.segments.append(Segment(segment[0], segment[1], data))
            self.size += len(data)

    def _read_bin(self, source: str | bytearray):
        if isinstance(source, str):
            with open(source, "rb") as f:
                self.bytes = bytearray(f.read())
        else:
            self.bytes = source
        self.size = len(self.bytes)

    def crc32(self):
        """
        Return the crc32 checksum of the firmware image

        Return:
            The firmware's CRC32, ready for comparison with the CRC
            returned by the ROM bootloader's COMMAND_CRC32
        """
        if self._crc32 is None:
            self._crc32 = binascii.crc32(bytearray(self.bytes)) & 0xFFFFFFFF

        return self._crc32

    def check_bootloader(self, offset: int) -> bool:
        """
        Check if the settings of bootloader in firmware image

        Parameters:
            offset -- The offset to the BL_CONFIG field in the CCFG area

        Return:
            True if the bootloader backdoor is enabled, False otherwise
        """
        data = self.segments[-1].bytes[offset : offset + 4]
        _LOGGER.debug("BL_CONFIG: 0x%08x" % struct.unpack("<I", data)[0])

        bl_enabled = data[0] == 0xC5 and data[3] == 0xC5
        bl_pin = data[1]
        active_low = not (data[2] & 0x01)

        if bl_enabled:
            _LOGGER.info("BL_PIN: %d, Active Low: %s", bl_pin, active_low)
        return bl_enabled



#!/usr/bin/env python3

# Copyright (c) 2014, Jelmer Tiete <jelmer@tiete.be>.
# Copyright (c) 2024, SMLIGHT <smartlight.email@gmail.com>
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import glob
import logging
import re

import coloredlogs
from tqdm.asyncio import tqdm

from . import __version__
from .command import Bootloader, CommandInterface
from .const import MAX_BLOCK_SIZE
from .device import CC26xx
from .exceptions import CliException
from .firmware import FirmwareFile

# Serial boot loader over UART for CC13xx / CC26xx
# Based on the info found in TI's swru333a.pdf (spma029.pdf)
#
# Bootloader only starts if no valid image is found or if boot loader
# backdoor is enabled.
# Make sure you don't lock yourself out!! (enable backdoor in your firmware)
# More info at https://github.com/JelmerT/cc2538-bsl


_LOGGER = logging.getLogger(__name__)


class CLI:
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog="smlight_cc_flasher")
        self.args = self.cli_setup()
        self.validate_args(self.args)

    def config_logging(self, log_level):
        coloredlogs.install(level=log_level, milliseconds=True)

    def validate_device(self, value):
        windows_pattern = r"^COM\d+$"
        linux_pattern = r"^/dev/tty(USB|ACM)\d+$"
        macos_pattern = r"^/dev/tty\.(usbserial|usbmodem)[A-Za-z0-9.-]+$"
        socket_pattern = r"^socket://[A-Za-z0-9.-]+:\d+$"

        if (
            re.match(windows_pattern, value)
            or re.match(linux_pattern, value)
            or re.match(macos_pattern, value)
            or re.match(socket_pattern, value)
        ):
            return value
        else:
            raise argparse.ArgumentTypeError(
                f"Invalid device format: {value}. "
                "Must be a serial port (e.g., COM7, /dev/ttyUSB0, "
                "/dev/tty.usbserial-1234) or a network socket (e.g., socket://host.local:9933)."
            )

    def cli_setup(self):
        parser = self.parser

        parser.add_argument("-q", action="store_true", help="Quiet")
        parser.add_argument("-V", action="store_true", help="Verbose")
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Force operation(s) which are not safe",
        )
        parser.add_argument(
            "-e", "--erase", action="store_true", help="Erase device before flashing"
        )
        parser.add_argument("-w", "--write", action="store_true", help="Write")
        parser.add_argument(
            "-v", "--verify", action="store_true", help="Verify (CRC32 check)"
        )
        parser.add_argument("-r", "--read", action="store_true", help="Read")
        parser.add_argument(
            "-l",
            "--len",
            type=int,
            default=0x80000,
            help="Length of read (default: 0x%(default)x)",
        )
        parser.add_argument("-o", "--output", help="Output file for read data")
        parser.add_argument(
            "-d",
            "--device",
            type=self.validate_device,
            help="Serial port (/dev/tty*, COM<#> or socket://host:port)",
        )
        parser.add_argument("--host", help="Hostname or IP address")
        parser.add_argument(
            "-p",
            "--port",
            default=6638,
            help="Port number for network socket (default: %(default)s)",
        )
        parser.add_argument(
            "-b",
            "--baud",
            type=int,
            default=500000,
            help="Baudrate (default: %(default)s)",
        )
        parser.add_argument("-a", "--address", type=int, help="Target address")
        parser.add_argument(
            "-i", "--ieee-address", help="Set the secondary 64 bit IEEE address"
        )
        parser.add_argument(
            "--bootloader-reset",
            default="generic",
            choices=["generic", "none", "generic2", "network"],
            help="Bootloader mode",
        )
        parser.add_argument(
            "--bootloader-active-high",
            action="store_true",
            help="Use active high signals to enter bootloader",
        )
        parser.add_argument(
            "--bootloader-invert-lines",
            action="store_true",
            help="Inverts the use of RTS and DTR to enter bootloader",
        )

        parser.add_argument(
            "--m33", action="store_true", help="Cortex-M33 cc13x4/cc26x4 devices"
        )
        parser.add_argument(
            "--disable-bootloader",
            action="store_true",
            help="After finishing, disable the bootloader",
        )

        parser.add_argument(
            "--version", action="version", version="%(prog)s " + __version__
        )
        parser.add_argument("file")

        self.args = parser.parse_args()
        return self.args

    def validate_args(self, args):
        if args.read and not args.output:
            self.parser.error("--output is required when --read is specified")

        if args.read and not args.write and args.verify:
            self.parser.error("Verify after read not implemented.")

        if args.len < 0:
            self.parser.error("Length must be positive but %d was provided" % args.len)

        if args.V:
            self.config_logging("DEBUG")
        elif args.q:
            self.config_logging("WARNING")
        else:
            self.config_logging("INFO")

        if not (args.device or args.host):
            self.auto_port()

        if not args.device and args.host:
            args.device = f"socket://{args.host}:{args.port}"

    def auto_port(self):
        """Try to find the port automatically."""
        ports = []
        # use serial.tools.list_ports.comports() instead?
        # Get a list of all USB-like names in /dev
        for name in [
            "ttyACM",
            "tty.usbserial",
            "ttyUSB",
            "tty.usbmodem",
            "tty.SLAB_USBtoUART",
        ]:
            ports.extend(glob.glob("/dev/%s*" % name))

        ports = sorted(ports)

        if ports:
            # Found something - take it
            self.args.device = ports[0]
        else:
            raise Exception("No serial port found.")

    def parse_ieee_address(self, inaddr: str) -> int:
        """Convert an entered IEEE address into an integer"""
        try:
            return int(inaddr, 16)
        except ValueError:
            # inaddr is not a hex string, look for other formats
            if ":" in inaddr:
                bytes = inaddr.split(":")
            elif "-" in inaddr:
                bytes = inaddr.split("-")
            if len(bytes) != 8:
                raise ValueError("Supplied IEEE address does not contain 8 bytes")
            addr = 0
            for i, b in zip(range(8), bytes):
                try:
                    addr += int(b, 16) << (56 - (i * 8))
                except ValueError:
                    raise ValueError("IEEE address contains invalid bytes")
            return addr


async def main():
    cli = CLI()
    args = cli.args
    firmware = None
    cmd = CommandInterface()
    await cmd.open(args.device, args.baud)

    _LOGGER.info("Opening port %s, baud %d", args.device, args.baud)

    if args.bootloader_reset != "none":
        bl = Bootloader(args.device, cmd.transport)

        if args.bootloader_reset or args.host:
            bl.set_mode(args.bootloader_reset, args.host)

        if args.bootloader_active_high or args.bootloader_invert_lines:
            bl.set_options(args.bootloader_active_high, args.bootloader_invert_lines)

        await bl.invoke_bootloader()

        _LOGGER.info("Connecting to target...")
        device = CC26xx(cmd, firmware, args.m33)
        await device.connect()

        if args.write or args.verify:
            _LOGGER.info("Reading data from %s", args.file)
            firmware = FirmwareFile(path=args.file)
            device.set_firmware(firmware)
            # force erase when using segmented writes
            if args.write and firmware.segments:
                args.erase = True

            bl_active = firmware.check_bootloader(
                device.bootloader_address - device.CCFG_BASE - device.CCFG_START
            )
            if not (bl_active or args.force):
                raise Exception(
                    "Bootloader not active in firmware to be flashed.",
                    "Use --force to override.",
                )

    if args.address:
        device.set_flash_start_addr(args.address)

    if args.read:
        await device.read(args.len, args.output)
        _LOGGER.info("Read done")

    if args.erase:
        await device.erase()

    if args.write:
        pbar = tqdm(total=device._firmware.size, unit="B", unit_scale=True)

        if logging.getLogger().getEffectiveLevel() < logging.INFO:
            pbar.disable = True

        await device.flash(
            progress_callback=lambda current, _: pbar.update(MAX_BLOCK_SIZE)
        )
        pbar.close()
        _LOGGER.info("Write done")

    if args.verify:
        _LOGGER.info("Verifying by comparing CRC32 calculations.")
        await device.verify()

    if args.ieee_address:
        ieee_addr = cli.parse_ieee_address(args.ieee_address)
        if await device.set_ieee_address(ieee_addr):
            _LOGGER.info("Set address done")
        else:
            raise CliException("Set address failed")

    if args.disable_bootloader:
        device.disable_bootloader(args.force)

    await cmd.cmdReset()
    await cmd.close()



# Copyright (c) 2024, SMLIGHT <smartlight.email@gmail.com>
# SPDX-License-Identifier: Apache-2.0

import asyncio
import logging
import re
import socket
import struct

import serial_asyncio

from .const import (
    COMMAND,
    COMMAND_HAS_DATA,
    COMMAND_RET,
    COMMAND_STRS,
    MAX_BLOCK_SIZE,
    RETURN_CMD_STRS,
)
from .exceptions import CmdException

_LOGGER = logging.getLogger(__name__)


class PinSetter:
    def __init__(self, instance, attr_name):
        self.instance = instance
        self.attr_name = attr_name
        self._webserial = hasattr(instance, "webserial")

    async def __call__(self, value=None):
        if not self._webserial:
            if value is not None:
                setattr(self.instance, self.attr_name, value)
            return getattr(self.instance, self.attr_name)
        else:
            if value is not None:
                _LOGGER.info("_%s", self.attr_name)
                setter = getattr(self.instance, f"set_{self.attr_name}")
                await setter(value)
                _LOGGER.info(
                    "%s: %s",
                    self.attr_name,
                    getattr(self.instance, f"_{self.attr_name}"),
                )
            return

    def get_instance(self):
        return self.instance


class Bootloader:
    """
    Automatically invoke the bootloader on a device.
    Use DTR/RTS to reset the device and enter the bootloader.
    For SLZB-06 use the network API to invoke the bootloader.
    """

    CONNECT_TIMEOUT = 1

    def __init__(self, device, transport) -> None:
        self._device = device
        self._serial = transport.serial
        self._dtr_active_high = False
        self._inverted = False
        self._generic = True
        self._generic2 = False
        self._smlight_net = False
        match = re.match(r"socket://([^:]+):([0-9]+)", device)
        if match:
            self._host = match.group(1)
            self._smlight_net = True

    def set_mode(self, mode: str, host: str | None = None) -> None:
        if host and not hasattr(self, "_host"):
            self._host = host
            self._smlight_net = True

        if mode == "generic2":
            self._generic2 = True
        elif mode == "network":
            self._smlight_net = True

        if self._generic2 or self._smlight_net:
            self._generic = False
        return

    def set_options(self, dtr_active_high: bool, inverted: bool) -> None:
        self._dtr_active_high = dtr_active_high
        self._inverted = inverted

    async def invoke_bootloader(self):
        if not ("socket" in self._device or self._smlight_net):
            if self._inverted:
                self.set_bootloader_pin = PinSetter(self._serial, "rts")
                self.set_reset_pin = PinSetter(self._serial, "dtr")
            else:
                self.set_bootloader_pin = PinSetter(self._serial, "dtr")
                self.set_reset_pin = PinSetter(self._serial, "rts")
            if self._generic2:
                await self.invoke_generic2()
            else:
                await self.invoke_generic()
        else:
            await self.invoke_smlight_net(self._host)

        await asyncio.sleep(0.1)

    async def invoke_generic2(self):
        _LOGGER.info("Activating generic2 BSL")
        # Dongles that use Flipflop logic for bsl.
        # Connection between RTS DTR and reset and IO15:
        # DTR  RTS  |  RST  IO15
        # 1     1   |   1    1
        # 0     0   |   1    1
        # 1     0   |   0    1
        # 0     1   |   1    0
        await self.set_bootloader_pin(False)
        await self.set_reset_pin(True)
        await asyncio.sleep(0.1)
        await self.set_bootloader_pin(True)
        await self.set_reset_pin(False)
        await asyncio.sleep(0.2)
        await self.set_bootloader_pin(False)

    async def invoke_generic(self):
        await self.set_bootloader_pin(True if not self._dtr_active_high else False)
        await self.set_reset_pin(True)
        await self.set_reset_pin(False)
        await asyncio.sleep(0.2)
        await self.set_bootloader_pin(False if not self._dtr_active_high else True)
        # await asyncio.sleep(0.5)

    async def invoke_smlight_net(self, host) -> bool:
        from pysmlight.const import Commands
        from pysmlight.web import Api2

        _LOGGER.info("Activating SMLIGHT BSL via network")
        async with Api2(host) as client:
            if await client.check_auth_needed():
                return False

            return await client.set_cmd(Commands.CMD_ZB_BSL)


class CommandInterface:
    ACK = b"\x00\xCC"
    NACK = b"\x00\x33"

    async def open(self, aport=None, abaudrate=500000):
        self.reader, self.writer = await serial_asyncio.open_serial_connection(
            url=aport, baudrate=abaudrate  # , timeout=1
        )
        self.webserial = hasattr(self.writer.transport.serial, "webserial")
        _LOGGER.info("Opened port %s, baud %d", aport, abaudrate)

        sock = self.writer.transport.get_extra_info("socket")
        if sock is not None:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    async def close(self):
        _LOGGER.info("Closing port")

        self.writer.close()
        await self.writer.wait_closed()

    @property
    def transport(self):
        return self.writer.transport

    async def _wait_for_ack(self, cmd: COMMAND, timeout=1):
        got = bytearray()
        info = f"{COMMAND_STRS[cmd.value]} (0x{cmd.value:02X})"
        while True:
            try:
                got += await asyncio.wait_for(self.reader.read(1), timeout)
            except asyncio.exceptions.TimeoutError as e:
                if cmd != COMMAND.SYNCH:
                    raise CmdException("Timeout waiting for ACK/NACK after '%s'", info)
                raise e

            if CommandInterface.ACK in got:
                _LOGGER.debug("Got %d additional bytes before ACK/NACK", (len(got) - 2))
                return True
            elif CommandInterface.NACK in got:
                _LOGGER.debug("Target replied with a NACK during %s", info)
                return False

    def _encode_addr(self, addr: int) -> bytes:
        """encode 32-bit integer as 4 bytes"""
        return struct.pack(">I", addr)

    def _decode_addr(self, addr: bytes) -> int:
        """decode bytes as 32-bit integer"""
        return struct.unpack(">I", addr)[0]

    async def _write(self, data):
        if isinstance(data, int):
            assert data < 256
            data = bytes([data])

        length = len(data)
        self.writer.write(data)
        await self.writer.drain()

        buffer_size = 0
        if not self.webserial:
            buffer_size = self.writer.transport.get_write_buffer_size()
        assert isinstance(data, (bytes, bytearray)), "Bad data type"
        _LOGGER.debug("*** Wrote %d/%d bytes", buffer_size, length)

    async def _sendCmd(
        self,
        cmd,
        *,
        addr: int | None = None,
        size: int | None = None,
        data: bytes | bytearray | None = None,
        extra: bytes | bytearray | None = None,
    ) -> None:
        cmd = bytes([cmd.value])

        packet = cmd

        if addr is not None:
            packet += self._encode_addr(addr)

        if size is not None:
            packet += self._encode_addr(size)

        if extra is not None:
            packet += extra

        if data:
            assert len(data) <= 252
            packet += data

        crc = bytes([sum(packet) & 0xFF])
        packet = crc + packet
        if cmd != b"\x55":
            packet = bytes([len(packet) + 1]) + packet

        await self._write(packet)

    async def _processCmd(self, cmd, *, addr=None, size=None, data=None, extra=None):
        _LOGGER.debug("*** %s command (0x%02x)", COMMAND_STRS[cmd.value], cmd.value)
        await self._sendCmd(cmd, addr=addr, size=size, data=data, extra=extra)

        if await self._wait_for_ack(cmd):
            if cmd == COMMAND.RESET or cmd == COMMAND.SYNCH:
                return True

            ret_data = None
            if cmd.value in COMMAND_HAS_DATA:
                ret_data = await self.receivePacket()
                if cmd == COMMAND.GET_STATUS:
                    return ret_data

            ret = await self.checkLastCmd()

            if ret and ret_data:
                return ret_data
            return ret
        return False

    async def receivePacket(self):
        hdr = await self.reader.readexactly(2)
        size, chks = hdr
        data = await self.reader.readexactly(size - 2)

        _LOGGER.debug("*** received %x bytes", size)
        if chks == sum(data) & 0xFF:
            await self._write(CommandInterface.ACK)
            return data if len(data) > 1 else data[0]
        else:
            await self._write(CommandInterface.NACK)
            raise CmdException("Received packet checksum error")

    async def sendSynch(self):
        cmd = COMMAND.SYNCH

        if not self.webserial:
            self.writer.transport.flush()

        return await self._processCmd(cmd)

    async def checkLastCmd(self):
        status = await self.cmdGetStatus()
        if not (status):
            raise CmdException("No response from target on status request.")

        if status == COMMAND_RET.SUCCESS.value:
            return True
        else:
            status_str = RETURN_CMD_STRS.get(status, None)
            if status_str is None:
                _LOGGER.error("Warning: unrecognized status returned 0x%x", status)
            else:
                _LOGGER.error("Target returned: 0x%x, %s", status, status_str)
            return False

    async def cmdPing(self):
        cmd = COMMAND.PING

        return await self._processCmd(cmd)

    async def cmdReset(self):
        cmd = COMMAND.RESET

        return await self._processCmd(cmd)

    async def cmdGetChipId(self):
        cmd = COMMAND.GET_CHIP_ID

        version = await self._processCmd(cmd)

        if version:
            assert len(version) == 4, f"Unreasonable chip id: {repr(version)}"
            _LOGGER.debug("Version 0x%s", "".join(f"{v:02X}" for v in version))
            chip_id = (version[2] << 8) | version[3]
            return chip_id
        else:
            raise CmdException("GetChipID (0x28) failed")

    async def cmdGetStatus(self):
        cmd = COMMAND.GET_STATUS

        return await self._processCmd(cmd)

    async def cmdEraseSector(self, addr):
        """cc26xx only"""
        cmd = COMMAND.SECTOR_ERASE

        return await self._processCmd(cmd, addr=addr)

    async def cmdBankErase(self):
        cmd = COMMAND.BANK_ERASE

        return await self._processCmd(cmd)

    async def cmdCRC32(self, addr, size):
        cmd = COMMAND.CRC32
        extra = self._encode_addr(0x00000000)
        crc = await self._processCmd(cmd, addr=addr, size=size, extra=extra)
        if crc:
            return self._decode_addr(crc)

    async def cmdCRC32Segment(self, firmware):
        crc32 = []
        for segment in firmware.segments:
            crc32.append(await self.cmdCRC32(segment.start, segment.size))
        return crc32

    async def cmdDownload(self, addr, size):
        cmd = COMMAND.DOWNLOAD

        if (size % 4) != 0:  # check for invalid data lengths
            raise CmdException(
                "Invalid data size: %d. Size must be a multiple of 4.", size
            )

        return await self._processCmd(cmd, addr=addr, size=size)

    async def cmdDownloadCRC32(self, addr, size, crc):
        cmd = COMMAND.DOWNLOAD_CRC

        if (size % 4) != 0:  # check for invalid data lengths
            raise CmdException(
                "Invalid data size: %d. Size must be a multiple of 4.", size
            )

        return await self._processCmd(
            cmd, addr=addr, size=size, extra=self._encode_addr(crc)
        )

    async def cmdSendData(self, data):
        cmd = COMMAND.SEND_DATA
        # data = bytearray(data)
        assert isinstance(data, (bytes, bytearray)), "Bad data type"
        return await self._processCmd(cmd, data=data)

    async def cmdMemRead(self, addr):
        cmd = COMMAND.MEMORY_READ
        data = await self._processCmd(cmd, addr=addr, extra=b"\x01\x01")
        return data

    async def cmdMemWrite(self, addr, data, width):
        """cc26xx"""
        if width != len(data):
            raise ValueError("width does not match len(data)")
        if width != 1 and width != 4:
            raise ValueError("width must be 1 or 4")

        cmd = COMMAND.MEMORY_WRITE

        return await self._processCmd(cmd, addr=addr, data=data, extra=b"\x01")

    async def arange(self, start, stop, step):
        for i in range(start, stop, step):
            yield i
            await asyncio.sleep(0.0)

    async def writeMemory(self, addr, data, *, progress_callback=None):
        length = len(data)
        # amount of data bytes transferred per packet (theory: max 252 + 3)
        pkt_size = MAX_BLOCK_SIZE
        empty_packet = bytearray((0xFF,) * pkt_size)

        offset = 0
        addr_set = False
        last_result = False

        data_slices = (
            data[i : i + pkt_size] async for i in self.arange(0, len(data), pkt_size)
        )

        if progress_callback is not None:
            progress_callback(0, len(data))

        async for packet in data_slices:
            if packet != empty_packet:
                if not addr_set:
                    # set starting address if not set
                    await self.cmdDownload(addr + offset, length - offset)
                    addr_set = True
                _LOGGER.debug(
                    f"Writing {pkt_size} bytes starting at address"
                    f"0x{addr + offset:08X}"
                )

                last_result = await self.cmdSendData(packet)
                if not last_result:
                    raise CmdException("Write memory failed")

                if progress_callback is not None:
                    progress_callback(offset, length)
            else:
                addr_set = False

            offset += pkt_size

        return last_result



# Copyright (c) 2024, SMLIGHT <smartlight.email@gmail.com>
# SPDX-License-Identifier: Apache-2.0

from asyncio.exceptions import TimeoutError
import logging
import math
import struct

from .command import CommandInterface
from .exceptions import DeviceException
from .firmware import FirmwareFile

_LOGGER = logging.getLogger(__name__)


class Chip:
    def __init__(self, command_interface, m33=False):
        self.command_interface = command_interface

        self.flash_start_addr = 0x00000000
        self.bootloader_dis_val = 0xFFFFFFC5
        self.page_size = 2048
        self.m33 = m33

    def page_align_up(self, value):
        return int(math.ceil(value / self.page_size) * self.page_size)

    def page_align_down(self, value):
        return int(math.floor(value / self.page_size) * self.page_size)

    def page_to_addr(self, pages):
        addresses = []
        for page in pages:
            addresses.append(int(self.flash_start_addr) + int(page) * self.page_size)
        return addresses

    def set_flash_start_addr(self, addr):
        self.flash_start_addr = addr

    def disable_bootloader(self, force):
        if not force:
            error_str = (
                "Disabling the bootloader will prevent you from "
                "using this script until you re-enable the "
                "bootloader using JTAG. Use --force if sure.",
            )
            _LOGGER.warning(error_str)
            raise DeviceException("Aborted.")

        pattern = struct.pack("<L", self.bootloader_dis_val)

        if self.command_interface.writeMemory(self.bootloader_address, pattern):
            _LOGGER.info("Set bootloader closed done")
        else:
            raise DeviceException("Set bootloader closed failed")


class CC26xx(Chip):
    # Class constants
    PROTO_MASK_BLE = 0x01
    PROTO_MASK_IEEE = 0x04
    PROTO_MASK_BOTH = 0x05

    def __init__(
        self,
        command_interface: CommandInterface,
        firmware: FirmwareFile | None = None,
        m33: bool = False,
    ):
        super().__init__(command_interface, m33)
        if firmware:
            self._firmware: FirmwareFile = firmware

        if self.m33:
            self.FCFG_BASE = 0x50000800  # P10
            self.CCFG_BASE = 0x50000000
            self.CCFG_OFFSET = 0x00
            self.CCFG_START = 0x00
            self.page_size = 2048
            self.FLASH_BASE = 0x58030000
        else:
            self.FCFG_BASE = 0x50001000
            self.CCFG_BASE = 0x50003000
            self.page_size = 4096
            self.CCFG_OFFSET = 0x1FB0
            self.CCFG_START = 0x1FA8
            self.FLASH_BASE = 0x40030000

    def set_firmware(self, file: FirmwareFile):
        self._firmware = file

    async def async_init(self):
        fcfg_user_id_offset = 0x294
        fcfg_icepick_id_offset = 0x318
        flash_size_offset = 0x2C
        ieee_address_primary_offset = 0x2F0  # FCFG1
        ieee_address_secondary_offset = 0x18  # CCFG
        bl_config_offset = 0x28
        self.misc_conf_1_offset = 0xA0  # FCFG1

        # Determine CC13xx vs CC26xx via ICEPICK_DEVICE_ID::WAFER_ID and store
        # PG revision
        device_id = await self.command_interface.cmdMemRead(
            self.FCFG_BASE + fcfg_icepick_id_offset
        )
        wafer_id = (
            ((device_id[3] & 0x0F) << 16) + (device_id[2] << 8) + (device_id[1] & 0xF0)
        ) >> 4
        pg_rev = (device_id[3] & 0xF0) >> 4

        # Read FCFG1_USER_ID to get the package and supported protocols
        user_id = await self.command_interface.cmdMemRead(
            self.FCFG_BASE + fcfg_user_id_offset
        )
        package = {
            0x00: "4x4mm",
            0x01: "5x5mm",
            0x02: "7x7mm",
            0x03: "Wafer",
            0x04: "2.7x2.7",
            0x05: "7x7mm Q1",
        }.get(user_id[2] & 0x07, "Unknown")

        protocols = user_id[1] >> 4

        # We can now detect the exact device
        _LOGGER.info(
            "pg_rev = {:0x}, protocols = {:0x}, wafer_id = 0x{:04x}".format(
                pg_rev, protocols, wafer_id
            ),
        )

        if wafer_id == 0xB99A:  # CC26x0
            chip = await self._identify_cc26xx(pg_rev, protocols)
        elif wafer_id == 0xB9BE:  # CC13x0
            chip = await self._identify_cc13xx(pg_rev, protocols)
        elif wafer_id == 0xBB41:  # CC13x2/CC26x2
            chip = await self._identify_cc13xx(pg_rev, protocols)
            self.page_size = 8192
        elif wafer_id == 0xBB77:  # CC13x2x7/CC26x2x7
            chip = await self._identify_cc13xx(pg_rev, protocols)
            self.page_size = 8192
        elif wafer_id == 0xBB78:  # CC13x4/CC27x4P
            chip = await self._identify_cc2674(pg_rev)
            self.page_size = 2048
        else:
            raise DeviceException("Unknown wafer_id: 0x%04x", wafer_id)

        flash_size = await self.command_interface.cmdMemRead(
            self.FLASH_BASE + flash_size_offset
        )
        if self.m33:
            self.size = (
                (((flash_size[1] & 0x03) << 1) + ((flash_size[0] & 0x80) >> 7))
                * 256
                * 1024
            )
        else:
            self.size = flash_size[0] * self.page_size

        self.bootloader_address = self.CCFG_BASE + self.CCFG_OFFSET + bl_config_offset
        self.addr_ieee_address_secondary = (
            self.CCFG_BASE + self.CCFG_OFFSET + ieee_address_secondary_offset
        )

        # Primary IEEE address. Stored with the MSB at the high address
        ieee_addr = await self.command_interface.cmdMemRead(
            self.FCFG_BASE + ieee_address_primary_offset + 4
        )
        ieee_addr = ieee_addr[::-1]
        ieee_addr2 = await self.command_interface.cmdMemRead(
            self.FCFG_BASE + ieee_address_primary_offset
        )
        ieee_addr += ieee_addr2[::-1]

        _LOGGER.info(
            f"{chip} ({package}): {self.size >> 10}KB Flash, "
            f"CCFG.BL_CONFIG at 0x{self.bootloader_address:08X}"
        )
        _LOGGER.info(
            "Primary IEEE Address: %s", ":".join(f"{x:02x}" for x in ieee_addr)
        )

    async def _identify_cc26xx(self, pg, protocols):
        chips_dict = {
            CC26xx.PROTO_MASK_IEEE: "CC2630",
            CC26xx.PROTO_MASK_BLE: "CC2640",
            CC26xx.PROTO_MASK_BOTH: "CC2650",
        }

        chip_str = chips_dict.get(protocols & CC26xx.PROTO_MASK_BOTH, "Unknown")

        if pg == 1:
            pg_str = "PG1.0"
        elif pg == 3:
            pg_str = "PG2.0"
        elif pg == 7:
            pg_str = "PG2.1"
        elif pg == 8 or pg == 0x0B:
            # CC26x0 PG2.2+ or CC26x0R2
            rev_minor = await self.command_interface.cmdMemRead(
                self.FCFG_BASE + self.misc_conf_1_offset
            )
            rev_minor = rev_minor[0]
            if rev_minor == 0xFF:
                rev_minor = 0x00

            if pg == 8:
                # CC26x0
                pg_str = f"PG2.{(2 + rev_minor):d}"
            elif pg == 0x0B:
                # HW revision R2, update Chip name
                chip_str += "R2"
                pg_str = "PG%d.%d" % (1 + (rev_minor // 10), rev_minor % 10)

        return f"{chip_str} {pg_str}"

    async def _identify_cc2674(self, pg):
        chip_str = "CC2674"
        if pg == 0:
            pg_str = "PG1"
        elif pg == 1:
            pg_str = "PG2"

        rev_minor = await self.command_interface.cmdMemRead(
            self.FCFG_BASE + self.misc_conf_1_offset
        )
        rev_minor = rev_minor[0]
        if rev_minor == 0xFF:
            rev_minor = 0x00

        pg_str = f"{pg_str}.{rev_minor:d}"

        return f"{chip_str} {pg_str}"

    async def _identify_cc13xx(self, pg, protocols):
        chip_str = "CC131x"
        if protocols & CC26xx.PROTO_MASK_IEEE == CC26xx.PROTO_MASK_IEEE:
            chip_str = "CC135x"

        if pg == 0:
            pg_str = "PG1.0"
        if pg == 1:
            pg_str = "PG1.1"
        elif pg == 2 or pg == 3:
            rev_minor = await self.command_interface.cmdMemRead(
                self.FCFG_BASE + self.misc_conf_1_offset
            )
            rev_minor = rev_minor[0]
            if rev_minor == 0xFF:
                rev_minor = 0x00
            pg_str = f"PG2.{rev_minor:d}"
        else:
            pg_str = "PG1.0"

        return f"{chip_str} {pg_str}"

    async def crc(self, address, size):
        return await self.command_interface.cmdCRC32(address, size)

    async def connect(self):
        try:
            await self.command_interface.sendSynch()
        except TimeoutError:
            raise DeviceException(
                "Can't connect to target. Ensure boot loader "
                "is started. (no answer on synch sequence)"
            )
        await self.async_init()

    async def erase(self):
        ret = await self.erase_bank()
        if ret and self.m33:
            ret = await self.erase_ccfg()
        if ret:
            _LOGGER.info("Erase done")
        else:
            raise DeviceException("Erase failed")

    async def erase_bank(self):
        _LOGGER.info("Erasing all main bank flash sectors")
        return await self.command_interface.cmdBankErase()

    async def erase_ccfg(self):
        _LOGGER.info("Erasing CCFG")
        return await self.command_interface.cmdEraseSector(self.CCFG_BASE)

    async def read_memory(self, addr):
        return await self.command_interface.cmdMemRead(addr)

    async def set_ieee_address(self, ieee_addr):
        formatted_addr = f"{':'.join(f'{b:02x}' for b in struct.pack('>Q', ieee_addr))}"
        _LOGGER.info("Setting IEEE address to %s", formatted_addr)
        ieee_addr_bytes = struct.pack("<Q", ieee_addr)

        return await self.command_interface.writeMemory(
            self.addr_ieee_address_secondary, ieee_addr_bytes
        )

    async def flash(self, progress_callback=None):
        if self._firmware.segments:
            for segment in self._firmware.segments:
                msg = (
                    f"Writing {segment.size} bytes starting at address "
                    f"0x{segment.start:08X}"
                )
                _LOGGER.debug(msg)

                await self.command_interface.writeMemory(
                    segment.start, segment.bytes, progress_callback=progress_callback
                )

        else:
            await self.command_interface.writeMemory(
                self.flash_start_addr,
                self._firmware.bytes,
                progress_callback=progress_callback,
            )

    async def read(self, length, output):
        # Round up to a 4-byte boundary
        length = (length + 3) & ~0x03

        _LOGGER.info(
            f"Reading {length} bytes starting at address 0x{self.flash_start_addr:x}",
        )
        with open(output, "wb") as f:
            for i in range(0, length >> 2):
                # reading 4 bytes at a time
                rdata = await self.read_memory(self.flash_start_addr + (i * 4))
                _LOGGER.debug(
                    " 0x%x: 0x%02x%02x%02x%02x"
                    % (
                        self.flash_start_addr + (i * 4),
                        rdata[0],
                        rdata[1],
                        rdata[2],
                        rdata[3],
                    )
                )
                f.write(rdata)

    async def verify(self):
        if self._firmware.segments:
            crc_list = await self.command_interface.cmdCRC32Segment(self._firmware)

            for idx in range(len(crc_list)):
                if crc_list[idx] == self._firmware.segments[idx].crc32:
                    _LOGGER.info(
                        f"Verified Segment {idx:d} (match: 0x{crc_list[idx]:08x})"
                    )
                else:
                    await self.command_interface.cmdReset()
                    raise Exception(
                        f"NO CRC32 match segment {idx:d}: Local = 0x{crc_list[idx]:x}, "
                        f"Target = 0x{self._firmware.segments[idx].crc32:x}"
                    )
        else:
            crc_local = self._firmware.crc32()
            crc_target = self.crc(self.flash_start_addr, len(self._firmware.bytes))

            if crc_local == crc_target:
                _LOGGER.info("Verified (match: 0x%08x)", crc_local)
            else:
                await self.command_interface.cmdReset()
                raise Exception(
                    "NO CRC32 match: Local = 0x{:x}, "
                    "Target = 0x{:x}".format(crc_local, crc_target)
                )



# Copyright (c) 2024, SMLIGHT <smartlight.email@gmail.com>
# SPDX-License-Identifier: Apache-2.0

""" Flasher class for api use via cc-web-tools or HA"""

import logging
from typing import Any, Callable, Optional

from .command import Bootloader, CommandInterface
from .device import CC26xx
from .firmware import FirmwareFile

LOG_LEVELS = [logging.INFO, logging.DEBUG]
VERBOSE = True
_LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=LOG_LEVELS[min(len(LOG_LEVELS) - 1, VERBOSE)])


class Flasher:
    def __init__(
        self,
        device: str,
        *,
        baudrate: int = 500000,
        m33: bool = False,
        bsl2: bool = False,
    ):
        self._device = device
        self._baudrate = baudrate
        self._m33 = m33
        self.bsl2 = bsl2
        self.command_interface = CommandInterface()
        self.firmware: FirmwareFile | None = None

    async def async_init(self, file: str | None = None):
        if file is not None:
            self.firmware = FirmwareFile(path=file)
        self.chip = CC26xx(self.command_interface, self.firmware, self._m33)
        await self.command_interface.open(self._device, self._baudrate)
        self.bootloader = Bootloader(self._device, self.command_interface.transport)
        if self.bsl2:
            self.bootloader.set_mode("generic2")

    async def connect(self):
        _LOGGER.debug("Activate bootloader")
        await self.bootloader.invoke_bootloader()
        await self.chip.connect()

    def set_firmware(self, file):
        self.chip.set_firmware(file)

    async def flash(
        self, progress_callback: Optional[Callable[[int, int], Any]] = None
    ):
        await self.chip.erase()
        await self.chip.flash(progress_callback=progress_callback)
        await self.chip.verify()
        await self.command_interface.cmdReset()
        await self.command_interface.close()
